"""
(c) Copyright 2015 Hexoskin
Permission to use, copy, modify, and distribute this software for any purpose with or without fee is hereby granted,
provided that the above copyright notice and this permission notice appear in all copies. The software is provided
"as is" and the author disclaims all warranties with regard to this software including all implied warranties of
merchantability and fitness. In no event shall the author be liable for any special, direct, indirect, or
consequential damages or any damages whatsoever resulting from loss of use, data or profits, whether in an action of
contract, negligence or other tortious action, arising out of or in connection with the use or performance of this
software.
"""

import csv
import json
import os
import struct
import wave

__version__ = '1.0.2'


def load_data(filepath, start=0):
    """
    Loads a  file according to the specified format. Time t=0 corresponds to
    the beginning of the record.
    :param str filepath:
    :param int start: an offset in second to add to the time of the data
    :return: np.array
    """
    data = []
    if os.path.exists(filepath + '.wav'):
        data = load_wave(filepath + '.wav', start)
    elif os.path.exists(filepath + '.csv'):
        data = load_csv(filepath + '.csv', start)
    return data


def load_csv(filepath, start=0):
    """
    Loads a csv file according to the specified format. Time t=0 corresponds to
    the beginning of the record.
    :param str filepath:
    :param int start: an offset in second to add to the time of the data
    :return: np.array
    """
    data = []
    file_specs = getspec(filepath)
    with open(filepath, 'r') as csvfile:
        reader_csv = csv.reader(csvfile, delimiter=',', quotechar='|')
        for row in reader_csv:
            try:
                try:
                    int(row[1])  # ensure it is an int otherwise is was already converted, don't change second column
                    data.append([float(row[0]) + start, float(row[1]) * file_specs['gain'] + file_specs['offset']])
                except ValueError:
                    data.append([float(row[0]) + start, float(row[1])])
            except ValueError:
                pass
    return data


def load_wave(filepath, start=0):
    """
    Loads a WAV file according to the specified format. Time t=0 corresponds to
    the beginning of the record.
    :param str filepath:
    :param int start: an offset in second to add to the time of the data
    :return:
    """
    #note data can be loaded with scipy.io.wavfileor numpy, but we keep here a decoding possible
    # with the basic python package
    file_specs = getspec(filepath)
    f = wave.open(filepath)
    if file_specs['signed']:
        yy = struct.unpack("h" * f.getnframes(), f.readframes(f.getnframes()))
    else:
        yy = struct.unpack("H" * f.getnframes(), f.readframes(f.getnframes()))
    f.close()
    freq_signal = f.getframerate()
    if 1 / freq_signal == file_specs['dt']:
        dt = file_specs['dt']
    elif freq_signal == file_specs['dt']:
        dt = file_specs['dt']
    else:
        raise ('invalid frequency {}'.format(filepath))
    return [(float(t) * dt + file_specs['dt_offset'] + start,
             float(y) * file_specs['gain'] + file_specs['offset']) for t, y in enumerate(yy)]


def get_files_specs(device_freq, as_dict=True):
    """
    Get all specs for a given frequency
    :param device_freq: device frequency 256 or 1000
    :return: table or dict of all available datatype
    """
    GAIN_ECG = 0.0064 if device_freq == 256 else 0.001
    OFFSET_ECG = -1360 * 0.0064 if device_freq == 256 else 0
    dt_e = 1 / 256 if device_freq == 256 else 1 / 250
    dt_r = 1 / 128 if device_freq == 256 else 1 / 125
    dt_a = 1 / 64 if device_freq == 256 else 1 / 50
    dt_o = 1 / 64 if device_freq == 256 else 1 / 75
    br_gain = 1 if device_freq == 256 else .1
    gain_resp = 13.28 if device_freq == 256 else 1
    tmp_gain = 1 if device_freq == 256 else 1 / 256
    keys_list = ['filename_in', 'filename_out', 'quality_signal_name', 'unit', 'dt', 'dt_offset',
                 'gain', 'offset', 'signed', 'round', 'datatype', 'is_wav']
    specs = [
        # filename_in, csv_name, quality_signal_name, unit, dt, dt_offset, gain, offset, signed, round, datatype, is_wav
        ["ECG_I", "ECG_I", "", "mV", dt_e, 0, GAIN_ECG, OFFSET_ECG, 1, 5, 4113, True],
        ["ECG_II", "ECG_II", "", "mV", dt_e, 0, GAIN_ECG, OFFSET_ECG, 1, 5, 4114, True],
        ["ECG_III", "ECG_III", "", "mV", dt_e, 0, GAIN_ECG, OFFSET_ECG, 1, 5, 4115, True],
        ["ECG_quality", "ECG_quality", "NONE", "na", 1, 1, 1, 0, 0, 0, 1002, True],
        ["respiration_thoracic", "respiration_thoracic", "", "na", dt_r, 0, 1, 0, 0, 5, 4129, True],
        ["respiration_abdominal", "respiration_abdominal", "", "na", dt_r, 0, 1, 0, 0, 5, 4130, True],
        ["acceleration_X", "acceleration_X", "", "G", dt_a, 0, 1 / 256, 0, 1, 8, 4145, True],
        ["acceleration_Y", "acceleration_Y", "", "G", dt_a, 0, 1 / 256, 0, 1, 8, 4146, True],
        ["acceleration_Z", "acceleration_Z", "", "G", dt_a, 0, 1 / 256, 0, 1, 8, 4147, True],
        ["ppgdo", "ppgdo", "", "na", dt_a, 0, 1, 0, 0, 5, 64, True],
        ["ppgDisc", "ppgDisc", "", "na", dt_a, 0, 1, 0, 0, 0, 64, True],
        ["ppgOOT", "ppgOOT", "", "na", dt_a, 0, 1, 0, 0, 5, 64, True],
        ["heart_rate", "heart_rate", "heart_rate_quality", "bpm", 1, 1, 1, 0, 0, 0, 19, True],
        ["energy_mifflin_keytel", "energy_mifflin_keytel", "", "watt", 1, 1, 1, 0, 0, 0, 42, True],
        ["heart_rate_quality", "heart_rate_quality", "NONE", "na", 1, 1, 1, 0, 0, 0, 1000, True],
        ["breathing_rate", "breathing_rate", "breathing_rate_quality", "rpm", 1, 1, br_gain, 0, 0, 0, 33, True],
        ["breathing_rate_quality", "breathing_rate_quality", "NONE", "na", 1, 1, 1, 0, 0, 0, 1001, True],
        ["minute_ventilation", "minute_ventilation_raw", "", "ml/min", 1, 1, gain_resp, 0, 0, 5, 36, True],
        ["tidal_volume", "tidal_volume_raw", "", "ml", 1, 1, gain_resp, 0, 0, 5, 37, True],
        ["minute_ventilation_adjusted", "minute_ventilation_adjusted", "", "ml/min", 1, 1, 10, 0, 0, 5, 38, True],
        ["tidal_volume_adjusted", "tidal_volume_adjusted", "", "ml", 1, 1, 1, 0, 0, 5, 39, True],
        ["minute_ventilation_cl", "minute_ventilation_adjusted", "", "ml/min", 1, 1, 10, 0, 0, 5, 38, True],
        ["tidal_volume_ml", "tidal_volume_adjusted", "", "ml", 1, 1, 1, 0, 0, 5, 39, True],
        ["activity", "activity", "", "G", 1, 1, 1 / 256, 0, 1, 8, 49, True],
        ["actigraphy", "actigraphy", "", "G", 1, 1, 1 / 256, 0, 1, 8, 54, True],
        ["cadence", "cadence", "", "spm", 1, 1, 1, 0, 0, 0, 53, True],
        ["temperature_celcius", "temperature_celcius", "", "Celcius", 1, 1, tmp_gain, 0, 1, 8, 81, True],
        ["temperature", "temperature", "", "Celcius", 1, 1, tmp_gain, 0, 1, 8, 81, True],
        ["RESP_TMPR", "RESP_TMPR", "", "Celcius", 1, 1, 0.01, 0, 1, 8, 338, True],
        ["systolic_pressure", "systolic_pressure", "", "mmHg", 1, 1, 1, 0, 0, 5, 98, True],
        ["systolic_pressure_adjusted", "systolic_pressure_adjusted", "", "mmHg", 1, 1, 1, 0, 0, 5, 99, True],
        ["SPO2", "SPO2", "SPO2_quality", "Percent", 1, 1, 1, 0, 0, 0, 66, True],
        ["SPO2_quality", "SPO2_quality", "NONE", "na", 1, 1, 1, 0, 0, 0, 1002, True],
        ["BATT_ST_CHANNEL_CHAR", "BATT_ST_CHANNEL_CHAR", "", "na", 1, 1, 1, 0, 0, 0, 247, True],
        ["TEMP_ST_CHANNEL_CHAR", "TEMP_ST_CHANNEL_CHAR", "", "na", 1, 1, 1, 0, 0, 0, 246, True],
        ["HRV_HF", "HRV_HF", "", "ms^2", 300, 300, 1 / 10, 0, 0, 8, 277, True],
        ["HRV_LF_normalized", "HRV_LF_normalized", "", "", 300, 300, 1 / 100, 0, 0, 8, 273, True],
        ["HRV_LF", "HRV_LF", "", "ms^2", 300, 300, 1 / 10, 0, 0, 8, 276, True],
        ["NN_over_RR", "NN_over_RR", "", "%", 300, 300, 1 / 100, 0, 0, 8, 274, True],
        ["ANN", "ANN", "", "s", 300, 300, 1 / device_freq / 16, 0, 0, 8, 271, True],
        ["SDNN", "SDNN", "", "s", 300, 300, 1 / device_freq / 16, 0, 0, 8, 272, True],
        ["HRV_triangular", "HRV_triangular", "", "na", 300, 300, 1 / 100, 0, 0, 5, 275, True],
        ["PPG", "PPG", "", "na", dt_o, 0, 1, 0, 0, 5, 64, True],
        ["QRS", "QRS_epoch", "", "na", 1, 0, 1, 0, 0, 0, 17, False],
        ["RR_interval", "RR_interval_epoch", "RR_interval_quality", "s", 1, 0, 1 / device_freq, 0, 0, 8, 18, False],
        ["inspiration", "inspiration_epoch", "", "na", 1, 0, 1, 0, 0, 0, 34, False],
        ["expiration", "expiration_epoch", "", "", 1, 0, 1, 0, 0, 0, 35, False],
        ["step", "step_epoch", "", "step", 1, 0, 1, 0, 0, 0, 52, False],
        ["PTT", "PTT_epoch", "", "s", 1, 0, 1 / device_freq, 0, 0, 0, 97, False],
        ["device_position", "device_position_epoch", "", "na", 1, 0, 1, 0, 0, 0, 269, False],
        ["sleep_position", "sleep_position_epoch", "", "na", 1, 0, 1, 0, 0, 0, 270, False],
        ["sleep_phase", "sleep_phase_epoch", "", "na", 1, 0, 1, 0, 0, 0, 280, False],
        ["NN_interval", "NN_interval_epoch", "", "s", 1, 0, 1 / device_freq, 0, 0, 8, 318, False],
        ["RR_interval_realigned", "RR_interval_realigned_epoch", "", "s", 1, 0, 1 / device_freq, 0, 0, 8, 319, False],
        ["RR_interval_quality", "RR_interval_quality_epoch", "", "na", 1, 0, 1 / device_freq, 0, 0, 0, 1004, False],
    ]
    if as_dict:
        return {vals[0]: {key: val for key, val in zip(keys_list, vals)} for vals in specs}
    else:
        return specs


def getspec(filepath):
    """
    return the specs of the given file
    :param str filepath:
    :return: dict
    """
    filename = os.path.splitext(os.path.split(filepath)[1])[0]
    info = load_info(os.path.dirname(filepath))
    device_freq = info['freq']

    try:
        return get_files_specs(device_freq)[filename]
    except ValueError:
        raise AssertionError('Not a valid filename {}'.format(filepath))


def load_info(directory):
    """
    load the info.json file
    :param str directory:
    :return: dict
     """
    with open(os.path.join(directory, "info.json"), "r") as f:
        info = json.load(f)
        info['freq'] = _check_freq(info)
    return info


def _check_freq(info):
    """Check the frequency of the data.
    This allow to load data from hexoskin (256 Hz) and Astroskin (1000 Hz).
    :param dict info:
    :return: frequency
    """

    root_start = info.get('start', info.get('start_timestamp', 0))
    if 1230768000 * 256 < root_start < 2524608000 * 256:  # 2009, 2050
        return 256
    elif 1230768000 * 1000 < root_start < 2524608000 * 1000:  # 2009, 2050
        return 1000
    else:
        raise AssertionError("start_timestamp ={}. out of range. Aborted.".format(root_start))
